/*
	PSEmu Plugin

	Plugin Type: GPU/PAD
	Plugin Name: Alleg

	Author: Linuzappz <linuzappz@hotmail.com>

*/
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#ifdef __LINUX__
#include <dirent.h>
#include <gtk/gtk.h>
#include <X11/keysym.h>
#endif

#include "GpuLocal.h"

#if defined (__LINUX__)
#include <signal.h>
typedef void* HWND;
#define CALLBACK
#elif defined (__WIN32__)
#include <winalleg.h>
#include <allegro/aintwin.h>
#define CALLBACK	__stdcall

HWND GpuWnd;
#endif

#include "PSEmu Plugin Defs.h"
#include "Cache.h"
#include "Prim.h"

// PPDK developer must change libraryName field and can change revision and build

const unsigned char version = PLUGIN_VERSION;	// do not touch - library for PSEmu 1.x

// it is up to developer but values must be in range 0-255

const unsigned char revision = 0;
const unsigned char build = 8;

//#define GPUONLY

#ifdef GPUONLY
char *libraryName = "Allegro Linuzappz Gpu Driver";
#else
char *libraryName = "Allegro Linuzappz Gpu/Pad Driver";
#endif

// driver dependant variables

void *_mangled_main_address;
BITMAP *vram = NULL;
unsigned char *vramtmp = NULL;
long MemSize;
long Conv24to15 = 0;
BITMAP *_bmp24 = NULL;
BITMAP *buffer = NULL;
BITMAP *drawvram = NULL;

int norender;
static int framecount;

int DspX; int DspY;
int DspW; int DspH;

GpuConf	conf;
unsigned int Mode;

unsigned int GpuStatus;
unsigned int GpuInfo[8];
int Rgb24;
short TPage;
short clipX;
short clipY;
short clipW;
short clipH;
short offsX;
short offsY;
int md;
int me;

// uncomment to log messages
//#define _LOG

int Log = 0;

static DispModeS OldDspMode;

static int ScreenOpened;
static int Initialized;

// for faster bgr->rgb conversion
unsigned short bgr2rgb[65536];

// internal GPU

static long GPUdataRet;
static unsigned long gpuData[64/4]; // Command Buffer == 64 bytes
static unsigned char gpuCommand;
static long gpuDataC;
static long gpuDataP;
static long DEnable;

long tWinX;
long tWinY;
long tWinW;
long tWinH;

long drawingLines;

long dispWidths[4] = {256,320,512,640};
long imageTransfer;

long textAddrY,textAddrX;

long textTP;
long textREST;
long textABR;

short imTYc,imTXc,imTY,imTX;
long imSize;
short imageX0,imageX1;
short imageY0,imageY1;

long newTextX0,newTextX1,newTextX2,newTextX3;
long newTextY0,newTextY1,newTextY2,newTextY3;

static void LoadConf();
static void SaveConf();

char dspText[256];
time_t dspT;
char fpsText[256];
char capText[256];

char* CALLBACK PSEgetLibName(void) {
    return libraryName;
}

unsigned long CALLBACK PSEgetLibType(void) {
#ifdef GPUONLY
    return PSE_LT_GPU;
#else
    return PSE_LT_GPU | PSE_LT_PAD;
#endif
}

unsigned long CALLBACK PSEgetLibVersion(void) {
    return version<<16|revision<<8|build;
}

void CALLBACK GPUmakeSnapshot(void) {
    FILE *bmpfile;
    BITMAP *bmp;
    char filename[256];
    char snapdir[8];
    unsigned int snapshotnr = 0;

    // increment snapshot value
    // get filename
    // try opening new snapshot file
   {
#ifdef __LINUX__
    DIR *dir;

    dir = opendir("snap");
    if (dir != NULL) {
	closedir(dir);
	strcpy(snapdir, "snap/");
    }
    else
#endif
	strcpy(snapdir, "");
   }

    for (;;) {
	snapshotnr++;
	sprintf(filename,"%ssnap%d.bmp", snapdir, snapshotnr);
	bmpfile = fopen(filename,"rb");
	if (bmpfile == NULL) break;
	fclose(bmpfile);
    }

    if (!conf.Stretch)
	bmp = create_sub_bitmap(screen, 0, 0, SCREEN_W, SCREEN_H);
    else
	bmp = create_sub_bitmap(screen, 0 + ((conf.DispW - DspW) / 2), 0 + ((conf.DispH - DspH) / 2), SCREEN_W, SCREEN_H);

    save_bmp(filename, bmp, NULL);
//    save_bmp(filename, vram, NULL);
}

long CALLBACK GPUinit() {
    if (Initialized) return 0;

    pslog ("GPUinit");

    // default config
    memset(&conf, 0, sizeof(conf));
    conf.AlterDsp = -1;
    conf.DispW = DEF_WIDTH;
    conf.DispH = DEF_HEIGHT;
    conf.CacheMode = 2;

    // force stretching if disp W < 640 or H < 480
    if ((conf.DispW < 640) || (conf.DispH < 480))
    	conf.Stretch = 1;

    // initialise device here ... 
    // possibly configure the color format, etc.

    fillLookupTables();
    GpuStatus = 0x74000000;

    memset(GpuInfo, 0, sizeof(GpuInfo));
    GpuInfo[7] = 2;

    DspX = DspY = 0;
    DspW = 640; DspH = 480;
    clipX = clipY = clipW = clipH = 0;

    norender = 0;

    Initialized = 1;

    return 0;
}

static int allegopen;
static int loadconf;

long CALLBACK GPUshutdown() {
    // shutdown device here....
    pslog ("GPUshutdown");

    if (!Initialized) return 0;
    if (ScreenOpened>0) {
	if (vramtmp != NULL) { free(vramtmp); vramtmp = NULL; }
	if (ScreenOpened == 1) CacheClose();
	ScreenOpened = 0;
    }

    Initialized = 0;
    if (allegopen) { allegro_exit(); allegopen = 0; }

    return 0;
}

#if defined (__LINUX__)
long GPUopen(unsigned long *Disp, char *pCapText, char *pCfgFile) {
#elif defined (__WIN32__)
long CALLBACK GPUopen(HWND hWnd) {
#endif

    pslog("GPUopen");
#if defined (__LINUX__)
    strncpy(capText, pCapText, 256);
#elif defined (__WIN32__)
    GpuWnd = hWnd;
#endif

    if (!allegopen) {
#ifdef __WIN32__
//	win_set_window(GpuWnd);
#endif
	if (allegro_init() == -1) {
    	    printf ("GpuAlleg: Error opening allegro\n");
	    return -1;
	}
	allegopen = 1;
    }
#ifdef GPUONLY
    install_keyboard();
#endif

    if (!loadconf) LoadConf();

    if (SetDspMode(DspW, DspH, Rgb24 ? 24 : 15) == -1) {
	GPUshutdown();
	return -1;
    }
    SetClip();

    ScreenOpened = 1;
    pslog ("GPUopen Ok");
    return 0;
}

long CALLBACK GPUclose() {
    int i;

    pslog ("GPUclose");

    if (ScreenOpened != 1) return 0;

    if (vram != NULL) {
        vramtmp = (unsigned char*)malloc(1024*512*2);
	if (vramtmp != NULL) {
	    acquire_bitmap(vram);
	    for (i=0;i<512;i++)
	        memcpy(vramtmp+i*1024*2, vram->line[i], 1024*2);
	    release_bitmap(vram);
	}
	CacheClose();
    }
    if (_bmp24 != NULL) { destroy_bitmap(_bmp24); _bmp24 = NULL; }
    if (buffer != NULL) { destroy_bitmap(buffer); buffer = NULL; }

#ifdef GPUONLY
	remove_keyboard();
#endif
//    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);

    OldDspMode.width = 0;
    OldDspMode.height = 0;
    OldDspMode.bpp = 0;

    ScreenOpened=2;

    return 0;
}

char *fpsmenu[] = 
    { "VSync","AlterDsp", "Stretch", "FrameSkip"  };
static int fpspos,fpssel;

#define FpsOption(var) \
    if (fpssel) var = 1 - var;\
    sprintf(tmp, "%s", var ? "on" : "off");

static char *AlterDspModes[4] = {
    "OFf",
    "Mode1",
    "Mode2",
    "Mode3"
};

void updateDisplay(void) {
    static time_t t1;
    static int nf,nft;
    int texty;

#ifdef __WIN32__
    if (key[KEY_DEL]) { conf.FpsCount = 1 - conf.FpsCount; key[KEY_DEL] = 0; }
    if (key[KEY_PGUP]) { if (fpspos < 3) fpspos++; key[KEY_PGUP] = 0; }
    if (key[KEY_PGDN]) { if (fpspos > 0) fpspos--; key[KEY_PGDN] = 0; }
    if (key[KEY_END]) { fpssel = 1; key[KEY_END] = 0; }
    if (key[KEY_ESC]) { GPUclose(); GPUshutdown(); exit(0); }
    if (key[KEY_F8]) GPUmakeSnapshot();
//    if (key[KEY_F5]) SendMessage(GpuWnd, WM_KEYUP, 0, VK_F5);
#endif

    if (DEnable) return;

    if (conf.FpsCount == 1) {
	if (time(NULL) > t1) {
	    nft = nf;
	    time(&t1);
	    nf = 0;
	}
	nf++;
    }

    if (conf.FrameSkip) {
	framecount++;
	if (conf.AlterDsp) {
	    if (framecount == 7) { norender = 1; framecount = 0; }
	    else if (norender) { norender = 0; return; }
	}
	else {
	    if (framecount <  2) return;
	    if (framecount == 7) { norender = 1; framecount = 0; }
	    if (framecount == 2) { norender = 0; return; }
	}
    }

    if (conf.AlterDsp == 2) {
	show_video_bitmap(drawvram == screen ? buffer : screen);
	drawvram = drawvram == screen ? buffer : screen;
    }
    else
    if (conf.AlterDsp) {
        UnsetClip();
	if (Conv24to15) {
	    blit(vram, _bmp24, DspX, DspY, 0, 0, DspW, DspH);
    	    if (vram == screen)
    	        stretch_blit(_bmp24, screen, 0, 0, DspW, DspH, 0, 512, conf.DispW, conf.DispH);
	    else
		stretch_blit(_bmp24, screen, 0, 0, DspW, DspH, 0, 0, conf.DispW, conf.DispH);
	}
	else {
    	    if (vram == screen) {
    		if (conf.Stretch)
    	    	    stretch_blit(vram, screen, DspX, DspY, DspW, DspH, 0, 512, conf.DispW, conf.DispH);
		else
    	    	    blit(vram, screen, DspX, DspY, 0 + ((conf.DispW - DspW) / 2), 512 + ((conf.DispH - DspH) / 2), DspW, DspH);
	    }
	    else {
		if (conf.Stretch)
    		    stretch_blit(vram, screen, DspX, DspY, DspW, DspH, 0, 0, conf.DispW, conf.DispH);
		else
    	    	    blit(vram, screen, DspX, DspY, 0 + ((conf.DispW - DspW) / 2), 0 + ((conf.DispH - DspH) / 2), DspW, DspH);
	    }
	}
	SetClip();
    }

    if (conf.FpsCount == 1) {
	char tmp[256];

	switch (fpspos) {
	    case 0:
		FpsOption(conf.VSync); fpssel = 0; break;
	    case 1:
		sprintf(tmp, "%s", AlterDspModes[conf.AlterDsp]);
/*		if (fpssel) {
		    loadconf=1;
		    GPUclose();
#if defined (__LINUX__)
		    GPUopen(NULL, capText, NULL);
#elif defined (__WIN32__)
		    GPUopen(GpuWnd);
#endif
		    loadconf=0;
		    fpssel = 0;
		}
*/		break;
	    case 2:
		if (!conf.AlterDsp) break;
		if ((conf.DispW < 640) || (conf.DispH < 480))
    		    { sprintf(tmp, "%s", conf.Stretch ? "on" : "off"); break; }
		FpsOption(conf.Stretch);
		if (fpssel) {
		    if (!conf.Stretch) clear(screen);
		    fpssel = 0;
		}
		break;
	    case 3:
		FpsOption(conf.FrameSkip);
		if (fpssel) { fpssel = 0; norender = 0; } break;
	}

	sprintf (fpsText,"%d FPS -- %s %s", nft, fpsmenu[fpspos], tmp);

	if (conf.AlterDsp && vram == screen) texty = 512 + 5;
	else texty = 5;

	UnsetClip();
	textout(screen, font, fpsText, 5, texty, 0xffffff);
	SetClip();
	texty += text_height(font);
    }
    else {
	if (conf.AlterDsp && vram == screen) texty = 512 + 5;
	else texty = 5;
    }

    if (*dspText) {
	if (time(NULL) > dspT) { *dspText = 0; dspT = 0; }
	UnsetClip();
	textout(screen, font, dspText, 5, texty, 0xffffff);
	SetClip();
    }

    if ((!conf.AlterDsp) && (conf.VSync==1)) vsync();

}


// update lace is called every VSync
void CALLBACK GPUupdateLace(void) {
    pslog ("update lace\n");
    if (ScreenOpened != 1) return;
    drawingLines^=1;
    updateDisplay();
}


// process read request from GPU status register
unsigned long CALLBACK GPUreadStatus(void) {
    // return the status of the GPU
    /*
    - gp1 - GPU Status
      Mask
    04000000 - 1:Idle 0:Busy
    10000000 - 1:Ready 0:Not ready , to receive commands.
    80000000 - GPU is drawing 1:Odd 0:Even lines, in interlaced mode.
    */
    //just'0x74000000' for now ! (idle, ready, even lines)
    unsigned long ret;

    if (drawingLines==1)
	ret = GpuStatus;
    else
	ret = GpuStatus|0x80000000;
    pslog ("readStatus %x",ret);
    return ret;
    //return 0x74000000;
}

// processes data send to GPU status register
// these are always single packet commands.
void CALLBACK GPUwriteStatus(unsigned long gdata) {
    pslog ("writeStatus cmd %x : %x\n",gdata>>24,gdata&0xffffff);
    switch(gdata>>24) {
	case 0x00:	/* Reset GPU */
	    GpuStatus = 0x14802000;
	    DspX = DspY = 0;
	    DspW = 256; DspH = 240;
	    SetDspMode(DspW, DspH, 15);
	    SetViewPort(DspX, DspY);
	    DEnable = 1;
	    Rgb24 = 0;
	    updateDisplay();
	    break;

	case 0x01:	/* Reset Command Buffer */
//	    memset(gpuData, 0, sizeof(gpuData));
	    break;

	case 0x02:	/* Reset IRQ */
	    break;

	case 0x03:	/* Display Enable */
	    GpuStatus = (GpuStatus & ~(1<<23)) | ((gdata&1)<<23);
	    DEnable = gdata&1;
//	    if (DEnable) SetDspMode(0, 0, 0);
//	    else SetDspMode(DspW, DspH, Rgb24 ? 24 : 15);
	    break;

	case 0x04:	/* DMA Setup */
//	    imageTransfer = gdata&3;
	    GpuStatus = (GpuStatus & ~(3<<29)) | ((gdata&3)<<29);
	    pslog("DMA Setup %x\n",gdata&3);
	    break;

	case 0x05:	/* Start Of Display Area */
	    DspY = (gdata>>10)&0x3ff;
	    DspX = gdata & 0x3ff;
	    SetViewPort(DspX, DspY);
	    break;

	case 0x06:	/* Horizontal Display Range */
//	    printf ("hdisp,%d,%d\n",gdata&0xfff,(gdata>>12)&0xfff);
//	    printf ("hdisp,%d,%d\n",(((gdata>>12)&0xfff)-(gdata&0xfff))/8,DspW);
	    {
		int i = (((gdata>>12)&0xfff)-(gdata&0xfff))/8;
		if ((Mode&0x43) == 0x40) DspW = 384;
	        else DspW = dispWidths[Mode&0x3];
		i = ((float)DspW / 320) * i;

		if (i != DspW) {
		    DspW = i;
	    	    if (SetDspMode(DspW, DspH, Rgb24 ? 24 : 15) == -1) {
			GPUshutdown();
			exit(1);
		    }
		}
	    }
	    break;

	case 0x07:	/* Vertical Display Range */
//	    printf ("vdisp,%d,%d\n",gdata&0x3ff,(gdata>>10)&0x3ff);
//	    printf ("vdisp,%d,%d\n",((gdata>>10)&0x3ff)-(gdata&0x3ff),DspH);
	    {
		int i = ((gdata>>10)&0x3ff) - (gdata&0x3ff);
		if (Mode&0x04) i *= 2;

		if (i != DspH) {
		    DspH = i;
	    	    if (SetDspMode(DspW, DspH, Rgb24 ? 24 : 15) == -1) {
			GPUshutdown();
			exit(1);
		    }
		}
	    }
	    break;

	case 0x08:	/* Display Mode */
	    if (gdata == Mode) break;
	    Mode = gdata;
	    if ((gdata&0x43) == 0x40) DspW = 384;
	    else DspW = dispWidths[gdata&0x3];
	    if (gdata&0x04) DspH = 480;
	    else DspH = 240;

	    Rgb24 = (gdata>>4)&0x1; // if 1 - TrueColor

	    //dispLaceNew = (gdata>>5)&0x01; // if 1 - Interlace

	    if (SetDspMode(DspW, DspH, Rgb24 ? 24 : 15) == -1) {
		GPUshutdown();
		exit(1);
	    }

//	    printf ("new mode %x\n",Mode);
//	    printf ("RGB24=%d,INTER=%d,REVERSEFLAG=%d\n",(gdata>>4)&1,(gdata>>5)&1,(gdata>>7)&1);
	    GpuStatus = (GpuStatus & ~(0x7f<<16)) | ((gdata&0x3f)<<17) | ((gdata&0x40)<<10);
	    break;

	case 0x10:	/* GPU Info */
	    switch (gdata&7) {
		case 3:	/* Draw area top left */
		case 4:	/* Draw area bottom right */
		case 5:	/* Draw Offset */
		case 7:	/* GPU Type */
		    GPUdataRet = GpuInfo[gdata&7];
		    break;
		default:
		    pslog ("GPU Info Not Implemented : %d",gdata&7);
	    }
	    break;

	default:
	    pslog ("unknown command gp1 %x",gdata);
	    break;
	}
}

unsigned long CALLBACK GPUreadData(void) {
    if(imageTransfer==3) {
	// image transfer from VRAM
	if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024)) {
	    acquire_bitmap(vram);
	    if (!Rgb24) {
		GPUdataRet=bgr2rgb[((unsigned short **)vram->line)[imTY][imTX]];
		GPUdataRet|=bgr2rgb[((unsigned short **)vram->line)[imTY][imTX+1]]<<16;
	    }
	    else {
		GPUdataRet=((unsigned short **)vram->line)[imTY][imTX];
		GPUdataRet|=((unsigned short **)vram->line)[imTY][imTX+1]<<16;
	    }
	    release_bitmap(vram);
	}
	imTX+=2;
        imTXc-=2;
	if(imTXc<=0) {
	    imTX=imageX0;
	    imTXc=imageX1;
	    imTYc--;
	    imTY++;
	}
	imSize--;          
	if(imSize <= 0) {
	    GpuStatus&=0xf7ffffff;
	    imageTransfer=0;
	}
    }
    pslog("GPUreadData %x\n",GPUdataRet);
    return GPUdataRet;
}

void CALLBACK GPUreadDataMem(unsigned long *pMem, int iSize) {
    int i;

    if(imageTransfer!=3) return;
    acquire_bitmap(vram);
    // image transfer from VRAM
    for (i=0;i<iSize;i++) {
	if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024)) {
	    if (!Rgb24) {
		pMem[i]=bgr2rgb[((unsigned short **)vram->line)[imTY][imTX]];
		pMem[i]|=bgr2rgb[((unsigned short **)vram->line)[imTY][imTX+1]]<<16;
	    }
	    else {
		pMem[i]=((unsigned short **)vram->line)[imTY][imTX];
		pMem[i]|=((unsigned short **)vram->line)[imTY][imTX+1]<<16;
	    }
	}
	imTX+=2;
	imTXc-=2;
	if(imTXc<=0) {
	    imTX=imageX0;
	    imTXc=imageX1;
	    imTYc--;
	    imTY++;
	}
	imSize--;          
	if(imSize <= 0) {
	    GpuStatus&=0xf7ffffff;
	    imageTransfer=0;
	}
    }
    release_bitmap(vram);
}

// processes data send to GPU data register
// user should not take care about that
void CALLBACK GPUwriteData(unsigned long gdata) {
    unsigned char command;

    GPUdataRet = gdata;
    //pslog("write %08x",gdata);
    if(imageTransfer == 2) {
	// image transfer to VRAM
	if (imTYc>0 && imTXc>0) {
	    acquire_bitmap(vram);
	    if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024)) {
		if (!Rgb24)
		    ((unsigned short **)vram->line)[imTY][imTX] = bgr2rgb[(unsigned short)gdata];
		else
		    ((unsigned short **)vram->line)[imTY][imTX] = (unsigned short)gdata;
	    }
	    imTX++;
	    imTXc--;
	    if(imTXc<=0) {
	        imTX=(imageX0);
	        imTXc=(imageX1);
	        imTYc--;
		imTY++;
	    }
	    if(imTYc <= 0) imageTransfer = 0;
     
	    if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024)) {
		if (!Rgb24)
		    ((unsigned short **)vram->line)[imTY][imTX] = bgr2rgb[(unsigned short)(gdata>>16)];
		else
		    ((unsigned short **)vram->line)[imTY][imTX] = (unsigned short)(gdata>>16);
	    }
	    release_bitmap(vram);

	    imTX++;
	    imTXc--;

	    if(imTXc<=0) {
	        imTX=(imageX0);
	        imTXc=(imageX1);
	        imTYc--;
		imTY++;
	    }
     
	    if(imTYc <= 0) imageTransfer = 0;
 
	    return;
	}
        else
	    imageTransfer = 0;
    }

    pslog("!GPU! DATA = %08x\n",gdata);
    if(gpuDataC == 0) {
	command = (unsigned char) (gdata>>24) & 0xff;
	if (primTableC[command]) {
	    gpuDataC = primTableC[command];
	    gpuCommand = command;
	    gpuData[0] = gdata;
	    gpuDataP = 1;
	}
	else  {
	    if (gdata) {
		pslog("!GPU! ERROR: COMMAND? %02x, %06x\n",command,gdata);
	    }
	    return;      
	}
    }
    else {
	gpuData[gpuDataP] = gdata;
	gpuDataP++;
    }
    if(gpuDataP == gpuDataC) {
	gpuDataC=gpuDataP=0;
	pslog ("Gpu(WriteData) Command %x",gpuCommand);
	primTableJ[gpuCommand]((unsigned char *)gpuData);
    }
}

void CALLBACK GPUwriteDataMem(unsigned long *pMem, int iSize) {
    unsigned short *pMemp = (unsigned short *)pMem;

    pslog ("writeDataMem %d , %d,%d,%d,%d\n",iSize,imTX,imTY,imTXc,imTYc);
    if (imageTransfer != 2) return;
    if ((imTX < 0) || (imTX + imTXc > 1024)) return;
    if ((imTY < 0) || (imTY + imTYc > 512)) return;
    acquire_bitmap(vram);
    while ((imTYc > 0) && (iSize > 0)) {
	if (!Rgb24) {
	    while (imTXc > 0) {
		((unsigned short **)vram->line)[imTY][imTX] = bgr2rgb[*pMemp++];
		imTXc--; imTX++;
		if (imTXc <= 0) break;
		((unsigned short **)vram->line)[imTY][imTX] = bgr2rgb[*pMemp++];
		imTXc--; imTX++;
	        iSize--;
	    }
	    imTX=imageX0; imTXc=imageX1;
	}
	else {
	    memcpy(&((unsigned short **)vram->line)[imTY][imTX], pMemp, imTXc*2);
	    iSize-=imTXc/2; pMemp+=imTXc;
	}
	imTYc--; imTY++;
    }
    if (imTYc <= 0) imageTransfer = 0;
    release_bitmap(vram);
}
/*
// this function will be removed soon
void CALLBACK GPUsetMode(unsigned long gdata) {
    imageTransfer = gdata;
}

// this function will be removed soon
long CALLBACK GPUgetMode(void) {
    return imageTransfer;
}
*/
long CALLBACK GPUdmaChain(unsigned long * baseAddrL, unsigned long addr) {
    // leave this function in peace :)
    // this is under high optimizing now
    // and however you dont need to change this
	
    unsigned long dmaMem;
    unsigned char * baseAddrB;
    unsigned char count;
    unsigned char cmd;

    pslog("GPUdmaChain");
    baseAddrB = (unsigned char*) baseAddrL;

    for (;;) {
	count = baseAddrB[addr+3];
	dmaMem = addr+4;
	while (count) {
	    if (imageTransfer == 2) {
		GPUwriteData(baseAddrL[dmaMem/4]);
		dmaMem+=4;
		count--;
		continue;
	    }
	    cmd = baseAddrB[dmaMem+3];
	    if (primTableC[cmd] == 0) {
		dmaMem+=4;
		count--;
	    }
	    else {
		pslog ("Gpu(DmaChain) Command %x : %x", cmd, baseAddrL[dmaMem/4]);
		primTableJ[cmd]((unsigned char*)baseAddrL + dmaMem);
		dmaMem+=(primTableC[cmd]*4);
		count-=primTableC[cmd];
	    }
	}
	addr = baseAddrL[addr/4]&0xffffff;
	if (addr <= 0 || addr == 0xffffff) break;
	addr&=0x7fffff;
    }

    return 0;
}

void CALLBACK GPUdisplayText(char *pText) {
    dspT = time(NULL) + 1;
    strncpy(dspText, pText, 256);
}

#ifdef __LINUX__

void CALLBACK GPUkeypressed(int key) {
    switch (key) {
	case XK_Delete: conf.FpsCount = 1 - conf.FpsCount; break;
	case XK_Page_Up: if (fpspos < 3) fpspos++; break;
	case XK_Page_Down: if (fpspos > 0) fpspos--; break;
	case XK_End: fpssel = 1; break;
	case XK_F8: GPUmakeSnapshot(); break;
    }
}

#endif

void fillLookupTables() {
    int i,r,g,b;

    i=0;
    for (b=0;b<32;b++)
	for (g=0;g<32;g++)
	    for (r=0;r<32;r++,i++)
		bgr2rgb[i] = MAKERGB15(r<<3, g<<3, b<<3);

    for (i=0;i<32768;i++)
	bgr2rgb[i+32768] = bgr2rgb[i] | 0x8000;
}

#define StartSetMode() \
	if (vram != NULL) { \
	    if (vramtmp == NULL) { \
	        vramtmp = (unsigned char*)malloc(1024*512*2); \
		if (vramtmp != NULL) { \
		    acquire_bitmap(vram); \
		    for (i=0;i<512;i++) \
			memcpy(vramtmp+i*1024*2, vram->line[i], 1024*2); \
		    release_bitmap(vram); \
		} \
	    } \
    	    CacheClose(); \
	}

#define EndSetMode() \
	if (CacheInit() == -1) return -1; \
	if (vram != NULL) { \
	    if (vramtmp != NULL) { \
		acquire_bitmap(vram); \
		for (i=0;i<512;i++) \
		    memcpy(vram->line[i], vramtmp+i*1024*2, 1024*2); \
		release_bitmap(vram); \
		free(vramtmp); \
		vramtmp = NULL;  \
	    } \
	}

int SetDspMode(int w, int h, int bpp) {
    int i;

//    printf ("setdspmode alterdsp = %d, mode = %dx%dx%d\n", conf.AlterDsp, w, h, bpp);

    if (conf.AlterDsp == -1) {
	set_color_depth(bpp);

  	if (set_gfx_mode(GFX_AUTODETECT, conf.DispW, conf.DispH, 0, 0) == -1) {
	    printf ("gpuAlleg: Error setting mode %d,%d: %s\n", conf.DispW, conf.DispH, allegro_error);
	    return -1;
	}
	AutoAlterDsp();
    }

    if (conf.AlterDsp == 3) {
	if (OldDspMode.bpp == bpp) return 0;
	StartSetMode();

	if ((vram != NULL) && (vram != screen)) destroy_bitmap(vram);

	set_color_depth(bpp);

	Conv24to15 = 0;
  	if (set_gfx_mode(GFX_AUTODETECT, conf.DispW, conf.DispH, 0, 0) == -1) {
	    set_color_depth(15);
  	    if (set_gfx_mode(GFX_AUTODETECT, conf.DispW, conf.DispH, 0, 0) == -1) {
		printf ("gpuAlleg: Error setting mode %d,%d (AlterDsp = %d): %s\n", conf.DispW, conf.DispH, conf.AlterDsp, allegro_error);
		return -1;
	    }
	    vram = create_bitmap_ex(24, 1024, 512);
	    if (vram == NULL) return -1;
	    if (conf.Stretch) {
		Conv24to15 = 1;
		if (_bmp24 != NULL) destroy_bitmap(_bmp24);
	        _bmp24 = create_system_bitmap(640, 512);
		if (_bmp24 == NULL) return -1;
	    }
	}
	else {
	    vram = create_system_bitmap(1024, 512);
	    if (vram == NULL) return -1;
	}
	drawvram = vram;

	MemSize = 0;

	EndSetMode();

#ifdef __LINUX__
	set_window_title(capText);
#endif
	OldDspMode.bpp = bpp;

	return 0;
    }

    if (conf.AlterDsp == 2) {
	if (OldDspMode.bpp == bpp) return 0;

	StartSetMode();

	set_color_depth(bpp);

	Conv24to15 = 0;
  	if (set_gfx_mode(GFX_AUTODETECT, conf.DispW, conf.DispH, 0, 0) == -1) {
	    set_color_depth(15);
  	    if (set_gfx_mode(GFX_AUTODETECT, conf.DispW, conf.DispH, 0, 0) == -1) {
		printf ("gpuAlleg: Error setting mode %d,%d (AlterDsp = %d): %s\n", conf.DispW, conf.DispH, conf.AlterDsp, allegro_error);
		printf ("gpuAlleg: Switching to Mode3\n");
		conf.AlterDsp = 3;
		return SetDspMode(w, h, bpp);
	    }
	    buffer = create_video_bitmap(conf.DispW, conf.DispH);
	    if (buffer == NULL) {
		conf.AlterDsp = 3;
		return SetDspMode(w, h, bpp);
	    }
	    if ((vram != NULL) && (vram != screen)) destroy_bitmap(vram);
	    vram = create_bitmap_ex(24, 1024, 512);
	    if (vram == NULL) return -1;
	    if (conf.Stretch) {
		Conv24to15 = 1;
		if (_bmp24 != NULL) destroy_bitmap(_bmp24);
	        _bmp24 = create_system_bitmap(640, 512);
		if (_bmp24 == NULL) return -1;
	    }
	}
	else {
	    buffer = create_video_bitmap(conf.DispW, conf.DispH);
	    if (buffer == NULL) {
		printf ("gpuAlleg: Error creating back buffer in mode2: %s\n", allegro_error);
		conf.AlterDsp = 3;
		return SetDspMode(w, h, bpp);
	    }
	    if ((vram != NULL) && (vram != screen)) destroy_bitmap(vram);
	    vram = create_system_bitmap(1024, 512);
	    if (vram == NULL) {
		printf ("gpuAlleg: Error creating system bitmap in mode2: %s\n", allegro_error);
		return -1;
	    }
	}
	drawvram = buffer;

	MemSize = gfx_driver->vid_mem;
	if (bpp == 15) MemSize-= conf.DispW*conf.DispH*2*2;
	else MemSize-= conf.DispW*conf.DispH*3*2;

	EndSetMode();

#ifdef __LINUX__
	set_window_title(capText);
#endif
	OldDspMode.bpp = bpp;

	return 0;
    }

    if (conf.AlterDsp == 1) {
	if (OldDspMode.bpp == bpp) return 0;

	StartSetMode();

	set_color_depth(bpp);

	Conv24to15 = 0;
	if (set_gfx_mode(GFX_AUTODETECT, conf.DispW, conf.DispH, 1024, 512 + conf.DispH) == -1) {
	    set_color_depth(15);
	    if (set_gfx_mode(GFX_AUTODETECT, conf.DispW, conf.DispH, 1024, 512 + conf.DispH) == -1) {
		printf ("gpuAlleg: Error setting mode %d,%d (AlterDsp = %d): %s\n", conf.DispW, conf.DispH, conf.AlterDsp, allegro_error);
		printf ("gpuAlleg: Switching to Mode2\n");
		conf.AlterDsp = 2;
		return SetDspMode(w, h, bpp);
	    }
	    vram = screen;
	    scroll_screen(0, 512);
	    Conv24to15 = 1;
	    if (_bmp24 != NULL) destroy_bitmap(_bmp24);
	    _bmp24 = create_system_bitmap(640, 512);
	    if (_bmp24 == NULL) return -1;
	}
	else {
	    vram = screen;
	    scroll_screen(0, 512);
	}
	drawvram = vram;

	MemSize = gfx_driver->vid_mem - 1024*512*2 - 1024*conf.DispH*2;

	EndSetMode();

#ifdef __LINUX__
	set_window_title(capText);
#endif
	OldDspMode.bpp = bpp;
	return 0;
    }

    if (conf.AlterDsp != 0) return -1;

    if ((OldDspMode.width == w) &&
	(OldDspMode.height == h) &&
	(OldDspMode.bpp == bpp))
	return 0;

    StartSetMode();

    set_color_depth(bpp);
    if (set_gfx_mode(GFX_AUTODETECT, w, h, 1024, 512) == -1) {
	printf ("gpuAlleg: Error setting mode %d,%d (AlterDsp = %d): %s\n", conf.DispW, conf.DispH, conf.AlterDsp, allegro_error);
	printf ("gpuAlleg: Switching to Mode1\n");
	conf.AlterDsp = 1;
	return SetDspMode(w, h, bpp);
    }
    vram = screen;
    drawvram = vram;

    MemSize = gfx_driver->vid_mem - 1024*512*2;

    EndSetMode();
#ifdef __LINUX__
    set_window_title(capText);
#endif

    OldDspMode.width = w;
    OldDspMode.height = h;
    OldDspMode.bpp = bpp;

    return 0;
}

int SetViewPort(int x, int y) {
    if (conf.AlterDsp) return 0;
    return scroll_screen(x, y);
}

void AutoAlterDsp() {
    conf.AlterDsp = 3;

    switch (gfx_driver->id) {
#if defined (__LINUX__)
	case GFX_XDGA:
	case GFX_XDGA_FULLSCREEN:
	case GFX_XDGA2:
	case GFX_VBEAF:
	    conf.AlterDsp = 1;
	    break;
#elif defined (__WIN32__)
        case GFX_DIRECTX:
        case GFX_DIRECTX_WIN:
        case GFX_DIRECTX_OVL:
	    conf.AlterDsp = 2;
#endif
    }
}

#if defined (__LINUX__)

GtkWidget *ConfDlg;
GtkWidget *EditW,*EditH;
GtkWidget *Combo;
GList *glist;

char *AlterDsplist[5] = {
    "Autodetect",
    "Off   - Display on vram, psx sized",
    "Mode1 - Display on vram, user sized",
    "Mode2 - Display on screen, user sized, drawing on screen",
    "Mode3 - Display on screen, user sized, drawing on vram"
};

static void OnOk() {
    int i;
    char *tmp;

    tmp = gtk_entry_get_text(GTK_ENTRY(EditW));
    conf.DispW = atoi(tmp);
    tmp = gtk_entry_get_text(GTK_ENTRY(EditH));
    conf.DispH = atoi(tmp);
    tmp = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(Combo)->entry));
    for (i=0;i<5;i++) if (!strcmp(tmp, AlterDsplist[i])) { conf.AlterDsp = i - 1; break; }

    gtk_widget_hide(ConfDlg);
    SaveConf();
}

static void OnCancel() {
    gtk_widget_hide(ConfDlg);
}

void OnVSync() {
    conf.VSync = 1 - conf.VSync;
}

void OnFrameSkip() {
    conf.FrameSkip = 1 - conf.FrameSkip;
}

void OnStretch() {
    conf.Stretch = 1 - conf.Stretch;
}

void OnFpsCount() {
    conf.FpsCount = 1 - conf.FpsCount;
}

#define CreateChkBtn(name,var,func) \
    Btn = gtk_check_button_new_with_label(name); \
    gtk_container_add(GTK_CONTAINER(Box2), Btn); \
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(Btn), var); \
    gtk_signal_connect (GTK_OBJECT(Btn), "clicked", GTK_SIGNAL_FUNC(func), NULL); \
    gtk_widget_show(Btn);

long GPUconfigure(void) {
    GtkWidget *Label;
    GtkWidget *Ok,*Cancel;
    GtkWidget *Box,*Box1,*Box2,*BBox;
    GtkWidget *Btn;
    char Text[8];
    int i;

    LoadConf();

    ConfDlg = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(ConfDlg, 380, 235);
    gtk_window_set_title(GTK_WINDOW(ConfDlg), "GPU Config Dialog");
    gtk_window_set_position(GTK_WINDOW(ConfDlg), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(ConfDlg), 10);

    Box = gtk_vbox_new(0, 0);
    gtk_container_add(GTK_CONTAINER(ConfDlg), Box);
    gtk_widget_show(Box);

    Box2 = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(Box), Box2, FALSE, FALSE, 10);
    gtk_widget_show(Box2);

    CreateChkBtn("Enable VSync",conf.VSync,OnVSync);

    Label = gtk_label_new("Select Alternate Display Mode:");
    gtk_box_pack_start(GTK_BOX(Box2), Label, FALSE, FALSE, 5);
    gtk_misc_set_alignment(GTK_MISC(Label), 0, 0);
    gtk_widget_show(Label);

    Combo = gtk_combo_new();
    gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(Combo)->entry), FALSE);
    glist = NULL;
    for (i=0;i<5;i++) glist = g_list_append(glist, AlterDsplist[i]);
    gtk_combo_set_popdown_strings(GTK_COMBO(Combo), glist);
    gtk_box_pack_start(GTK_BOX(Box2), Combo, FALSE, FALSE, 0);
    gtk_widget_set_usize(Combo, 360, 20);
    gtk_widget_show(Combo);
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(Combo)->entry), AlterDsplist[conf.AlterDsp + 1]);

    Box2 = gtk_vbox_new(0, 0);
    gtk_box_pack_start(GTK_BOX(Box), Box2, FALSE, FALSE, 0);
    gtk_widget_show(Box2);

    CreateChkBtn("Use Stretch",conf.Stretch,OnStretch);

    Box1 = gtk_hbox_new(0, 0);
    gtk_box_pack_start(GTK_BOX(Box), Box1, FALSE, FALSE, 5);
    gtk_widget_show(Box1);

    Label = gtk_label_new("Select Display Resolution:");
    gtk_box_pack_start(GTK_BOX(Box1), Label, FALSE, FALSE, 0);
    gtk_widget_show(Label);

    EditW = gtk_entry_new();
    sprintf(Text,"%d",conf.DispW);
    gtk_entry_set_text(GTK_ENTRY(EditW), Text);
    gtk_box_pack_start(GTK_BOX(Box1), EditW, FALSE, FALSE, 5);
    gtk_widget_set_usize(EditW, 40, 22);
    gtk_widget_show(EditW);

    EditH = gtk_entry_new();
    sprintf(Text,"%d",conf.DispH);
    gtk_entry_set_text(GTK_ENTRY(EditH), Text);
    gtk_box_pack_start(GTK_BOX(Box1), EditH, FALSE, FALSE, 5);
    gtk_widget_set_usize(EditH, 40, 22);
    gtk_widget_show(EditH);

    Box2 = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(Box), Box2, FALSE, FALSE, 5);
    gtk_widget_show(Box2);

    CreateChkBtn("Display Frame Counter",conf.FpsCount,OnFpsCount);
    CreateChkBtn("Use FrameSkip",conf.FrameSkip,OnFrameSkip);

    BBox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box), BBox, FALSE, FALSE, 5);
    gtk_widget_show(BBox);

    Ok = gtk_button_new_with_label("Ok");
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnOk), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Ok);
    GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
    gtk_widget_show(Ok);

    Cancel = gtk_button_new_with_label("Cancel");
    gtk_signal_connect(GTK_OBJECT(Cancel), "clicked", GTK_SIGNAL_FUNC(OnCancel), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Cancel);
    GTK_WIDGET_SET_FLAGS(Cancel, GTK_CAN_DEFAULT);
    gtk_widget_show(Cancel);

    gtk_widget_show(ConfDlg);

    return 0;
}

GtkWidget *AboutDlg;

static void OnAboutOk(GtkMenuItem *menuitem, gpointer userdata) {
    gtk_widget_hide(AboutDlg);
}

#define ABOUT_TEXT \
    "Gpu Driver for Allegro.\nBased on Duddie's Soft Gpu Driver.\nAuthor: Linuzappz\n" \
    "Thanks to:\n" \
    "Bobbi, for support\nPete Bernert, help and ideas\nFrodo Baggins, I used some code from the\n unreleased allegro plugin for fpse\n" \
    "Contact:\n" \
    "Send any bug, comment or help to:\n<linuzappz@pcsx.net>"


void GPUabout(void) {
    GtkWidget *Label;
    GtkWidget *Ok;
    GtkWidget *Box,*BBox;
    char AboutText[2048];

    strcpy (AboutText, ABOUT_TEXT);

    AboutDlg = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(AboutDlg, 300, 230);
    gtk_window_set_title(GTK_WINDOW(AboutDlg), "CDR About Dialog");
    gtk_window_set_position(GTK_WINDOW(AboutDlg), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(AboutDlg), 10);

    Box = gtk_vbox_new(0, 0);
    gtk_container_add(GTK_CONTAINER(AboutDlg), Box);
    gtk_widget_show(Box);

    Label = gtk_label_new(AboutText);
    gtk_box_pack_start(GTK_BOX(Box), Label, FALSE, FALSE, 0);
    gtk_widget_show(Label);

    BBox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box), BBox, FALSE, FALSE, 0);
    gtk_widget_show(BBox);

    Ok = gtk_button_new_with_label("Ok");
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnAboutOk), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Ok);
    GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
    gtk_widget_show(Ok);

    gtk_widget_show(AboutDlg);
}

#endif

long CALLBACK GPUtest(void) {
    return 0;
}

#ifdef __WIN32__
#define QueryKeyV(s, name, var) \
    size = s; \
    RegQueryValueEx(myKey, name, 0, &type, (LPBYTE) var, &size);
#endif

static void LoadConf() {
#if defined (__LINUX__)
    FILE *f;
    char cfg[255];

    sprintf (cfg,"%s/gpuAlleg.cfg",getenv("HOME"));

    f = fopen(cfg,"r");
    if (f == NULL) return;
    fscanf(f, "VSync = %d\n", &conf.VSync);
    fscanf(f, "AlterDsp = %d\n", &conf.AlterDsp);
    fscanf(f, "Stretch = %d\n", &conf.Stretch);
    fscanf(f, "DispW = %d\n", &conf.DispW);
    fscanf(f, "DispH = %d\n", &conf.DispH);
    fscanf(f, "FpsCount = %d\n", &conf.FpsCount);
    fscanf(f, "FrameSkip = %d\n", &conf.FrameSkip);
    fclose(f);
#elif defined (__WIN32__)
    HKEY myKey;
    DWORD type;
    DWORD size;

    if (RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Vision Thing\\PSEmu Pro\\GPU\\Alleg-LinuzAppz",0,KEY_ALL_ACCESS,&myKey)==ERROR_SUCCESS) {
	QueryKeyV(sizeof(DWORD), "FpsCount", &conf.FpsCount);
	QueryKeyV(sizeof(DWORD), "AlterDsp", &conf.AlterDsp);
	QueryKeyV(sizeof(DWORD), "FrameSkip", &conf.FrameSkip);
	QueryKeyV(sizeof(DWORD), "VSync", &conf.VSync);
	QueryKeyV(sizeof(DWORD), "DispW", &conf.DispW);
	QueryKeyV(sizeof(DWORD), "DispH", &conf.DispH);
	QueryKeyV(sizeof(DWORD), "Stretch", &conf.Stretch);
	RegCloseKey(myKey);
    }
#endif
}

static void SaveConf() {
#ifdef __LINUX__
    FILE *f;
    char cfg[255];

    sprintf (cfg,"%s/gpuAlleg.cfg",getenv("HOME"));

    f = fopen(cfg,"w");
    if (f == NULL) return;
    fprintf(f, "VSync = %d\n", conf.VSync);
    fprintf(f, "AlterDsp = %d\n", conf.AlterDsp);
    fprintf(f, "Stretch = %d\n", conf.Stretch);
    fprintf(f, "DispW = %d\n", conf.DispW);
    fprintf(f, "DispH = %d\n", conf.DispH);
    fprintf(f, "FpsCount = %d\n", conf.FpsCount);
    fprintf(f, "FrameSkip = %d\n", conf.FrameSkip);
    fclose(f);
#endif
}

void pslog(char *fmt, ...) {
#ifdef _LOG
    va_list list;
    static FILE *out=NULL;
    static char tmpbuf[128], tmpbufold[128];
    static int repeatcount;

    if (!Log) return;

    if (out == NULL) {
	out = fopen("log.txt","w");
	if (out == NULL) return;
    }

    va_start(list,fmt);
    vsprintf(tmpbuf, fmt, list);
    va_end(list);

    // don't log the same line
    if(strcmp(tmpbufold,tmpbuf)==0)
	repeatcount++;
    else {
	strcpy(tmpbufold,tmpbuf);
	if(repeatcount>0) {
	    fprintf(out,"Last msg repeated %d times\n",repeatcount);
	    repeatcount = 0;
	}
	if (tmpbuf[strlen(tmpbuf) - 1]=='\n')
	    fprintf(out,"%s",tmpbuf);
	else
	    fprintf(out,"%s\n",tmpbuf);
	fflush(out);
    }
#endif
}
