#include <stdio.h>
#include <stdlib.h>
#include "GpuLocal.h"
#include "Cache.h"


static Cache TexCache[CACHE_SIZE];
static int CacheAvail;
static int count;
static int cacheopen;

#define MAKE_TEX(xx,yy,ww,hh) \
    switch (tp) { \
	case 0: \
	    vramp = &((unsigned short **)vram->line)[cluy][clux]; \
	    for (x=0;x<16;x++) { \
		c = vramp[x]; \
		if ((c&0x7fff) == 0) clut[x] = MASK_COLOR_15; \
		else clut[x] = c; \
	    } \
	    for (y=yy;y<hh;y++) { \
		vramp = &((unsigned short **)vram->line)[y+texy][texx]; \
		texp = ((unsigned short **)TexCache[i].tex->line)[y]; \
		for (x=xx;x<(ww/4);x++) { \
		    c = bgr2rgb[vramp[x]]; \
		    *texp++ = clut[c&0xf]; \
		    *texp++ = clut[(c>>4)&0xf]; \
		    *texp++ = clut[(c>>8)&0xf]; \
		    *texp++ = clut[(c>>12)&0xf]; \
		} \
	    } \
	    break; \
	case 1: \
	    vramp = &((unsigned short **)vram->line)[cluy][clux]; \
	    for (x=0;x<256;x++) { \
		c = vramp[x]; \
		if ((c&0x7fff) == 0) clut[x] = MASK_COLOR_15; \
		else clut[x] = c; \
	    } \
	    for (y=yy;y<hh;y++) { \
		vramp = &((unsigned short **)vram->line)[y+texy][texx]; \
		texp = ((unsigned short **)TexCache[i].tex->line)[y]; \
		for (x=xx;x<(ww/2);x++) { \
		    c = bgr2rgb[vramp[x]]; \
		    *texp++ = clut[c&0xff]; \
		    *texp++ = clut[(c>>8)&0xff]; \
		} \
	    } \
	    break; \
	case 3: \
	case 2: \
	    for (y=yy;y<hh;y++) { \
		vramp = &((unsigned short **)vram->line)[y+texy][texx]; \
		texp = ((unsigned short **)TexCache[i].tex->line)[y]; \
		for (x=xx;x<ww;x++) { \
		    c = vramp[x]; \
		    if ((c&0x7fff) == 0) *texp++ = MASK_COLOR_15; \
		    else *texp++ = c; \
		} \
	    } \
	    break; \
    }


int CacheInit() {
    int i;

    if (cacheopen == 1) return 0;

    CacheAvail = 0;
    while (MemSize > 256*256*2) {
	CacheAvail++;
	MemSize-=256*256*2;
    }

    for (i=0;i<CACHE_SIZE;i++) {
	if (!CacheAvail) {
	    TexCache[i].tex = create_system_bitmap(256, 256);
	    if (TexCache[i].tex == NULL) return -1;
	}
	else {
	    CacheAvail--;
	    TexCache[i].tex = create_video_bitmap(256, 256);
	    if (TexCache[i].tex == NULL) {
		CacheAvail = 0;
	        TexCache[i].tex = create_system_bitmap(256, 256);
		if (TexCache[i].tex == NULL) return -1;
	    }
	}
	set_clip(TexCache[i].tex, 0, 0, 0, 0);
	TexCache[i].id = -1;
	TexCache[i].idold = -1;
	TexCache[i].u0 = 0;
	TexCache[i].v0 = 0;
	TexCache[i].u1 = 0;
	TexCache[i].v1 = 0;
    }

    cacheopen = 1;
    return 0;
}

Cache* CacheGet(unsigned long id, int u0, int v0, int u1, int v1) {
    int i=0,x,y,c;
    unsigned short clut[256];
    unsigned short *vramp;
    unsigned short *texp;
    short clux, cluy;
    short texx, texy;
    short tp;

    if (conf.CacheMode == 2)
	for (i=0;i<CACHE_SIZE;i++) {
	    if (id == TexCache[i].id) {
		return &TexCache[i];
	    }
	}
    else if (conf.CacheMode == 1) {
	for (i=0;i<CACHE_SIZE;i++) {
	    if (id == TexCache[i].id) {
		if ((u0 >= TexCache[i].u0) && (u1 <= TexCache[i].u1) &&
		    (v0 >= TexCache[i].v0) && (v1 <= TexCache[i].v1)) {
		    return &TexCache[i];
		}
		else {
		    pslog ("resizing %d,%d %d,%d : %d,%d %d,%d\n",u0,u1,v0,v1,TexCache[i].u0,TexCache[i].u1,TexCache[i].v0,TexCache[i].v1);
		    u0 = MIN(u0, TexCache[i].u0);
		    v0 = MIN(v0, TexCache[i].v0);
		    u1 = MAX(u1, TexCache[i].u1);
		    v1 = MAX(v1, TexCache[i].v1);
		    break;
		}
	    }
	}
    }

    // not found, make it

    pslog ("making cache tex id = %x\n",id);

    // maybe it was just updated, so use the old texture page
    if (i == CACHE_SIZE) {
        for (i=0;i<CACHE_SIZE;i++) {
	    if (id == TexCache[i].idold) break;
        }
	if (i == CACHE_SIZE) {
	    if (++count == CACHE_SIZE) count = 0;
	    i = count;
	}
    }

    TexCache[i].id = id;

    texx = (short)((id&0xf)<<6);
    texy = id&0x10 ? 256 : 0;
    clux = (short)((id>>12) & 0x3f0);
    cluy = (short)((id>>22) & 0x1ff);
    tp = (short)((id>>7)&3);

    acquire_bitmap(vram);
    acquire_bitmap(TexCache[i].tex);

    if (conf.CacheMode == 1) {
	pslog("tex[%d] making from %d,%d to %d,%d\n",i,u0,v0,u1,v1);
	MAKE_TEX(u0, v0, u1, v1);

	TexCache[i].u0 = u0;
	TexCache[i].u1 = u1;
	TexCache[i].v0 = v0;
	TexCache[i].v1 = v1;
    }
    else // make full texture
	MAKE_TEX(0, 0, 256, 256);

    release_bitmap(vram);
    release_bitmap(TexCache[i].tex);
/*    {
    char file[256];
    sprintf(file,"cache%x.bmp",id);
    save_bmp(file, TexCache[i].tex, NULL);
    }*/
    return &TexCache[i];
}

static short tpt[4] = {64, 128, 256, 256};

void CacheChanged(int x0, int y0, int x1, int y1) {
    int i;
    short tx0,ty0,tx1,ty1;
    short tp;

    for (i=0; i<CACHE_SIZE; i++) {
	tx0 = (short)((TexCache[i].id&0xf)<<6);
	ty0 = TexCache[i].id&0x10 ? 256 : 0;
	tp = (short)((TexCache[i].id>>7)&3);
	tx1 = tx0 + tpt[tp];
	ty1 = ty0 + 256;

	// is the changed vram inside the cache?
	// if so clear it
	if (((tx0 > x0) && (tx1 < x0)) || ((tx0 < x1) && (tx1 > x1)))
	    if (((ty0 > y0) && (ty1 < y0)) || ((ty0 < y1) && (ty1 > y1))) {
		TexCache[i].idold = TexCache[i].id;
		TexCache[i].id = -1;
		TexCache[i].u0 = 0;
		TexCache[i].v0 = 0;
		TexCache[i].u1 = 0;
		TexCache[i].v1 = 0;
	    }
    }
}

void CacheClear() {
    int i;

    for (i=0; i<CACHE_SIZE; i++) {
	TexCache[i].id = -1;
    }
}

void CacheClose() {
    int i;

    if (!cacheopen) return;

    for (i=0;i<CACHE_SIZE;i++) {
	if (TexCache[i].tex != NULL) destroy_bitmap(TexCache[i].tex);
	TexCache[i].id = -1;
    }

    cacheopen = 0;
}
