/*
 * Pad for Psemu Pro like Emulators
 *
 * By: linuzappz <linuzappz@hotmail.com>
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include <allegro.h>
#if defined (__LINUX__)
#include <sys/time.h>
#include <unistd.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <X11/keysym.h>

#define CALLBACK
typedef void* HWND;
#elif defined (__WIN32__)
#include <winalleg.h>

#define CALLBACK	__stdcall
#endif
#include <allegro.h>


#include "PSEmu Plugin Defs.h"

extern char *libraryName;
extern unsigned char build;
extern unsigned char revision;

void LoadConf();
void SaveConf();

unsigned short Pad1[16] = {
    KEY_SPACE,	// Select
    0,
    0,
    KEY_ENTER,	// Start
    KEY_UP,	// Up
    KEY_RIGHT,	// Right
    KEY_DOWN,	// Down
    KEY_LEFT,	// Left
    KEY_1,	// L2
    KEY_3,	// R2
    KEY_Q,	// L1
    KEY_E,	// R1
    KEY_W,	// Triangle
    KEY_D,	// Circle
    KEY_X,	// Cross
    KEY_A	// Square
};

unsigned short Pad2[16] = {
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0
};

long padflags=0;

long CALLBACK PADinit(long flags) {
    padflags|=flags;
    return 0;
}

long CALLBACK PADshutdown(void) {
    return 0;
}

int PadOpened = 0;

#if defined (__LINUX__)
long PADopen(unsigned long *Disp) {
#elif defined (__WIN32__)
long CALLBACK PADopen(HWND hWnd) {
#endif
    LoadConf();
    PadOpened = 1;
    install_keyboard();
    set_keyboard_rate(0, 0);
    return 0;
}

long CALLBACK PADclose(void) {
    remove_keyboard();
    PadOpened = 0;
    return 0;
}

long CALLBACK PADquery(void) {
    return 3; // both pads
}

typedef struct
{
    unsigned char controllerType;
    unsigned short buttonStatus;
    unsigned char rightJoyX, rightJoyY, leftJoyX, leftJoyY;
    unsigned char moveX, moveY;
    unsigned char reserved[91];
} PadDataS;

long CALLBACK PADreadPort1(PadDataS *pad) {
    static unsigned short Stat = 0xffff;
    int i;

    if (!(padflags & 1)) return 0xffff;

    for (i=0;i<16;i++) {
	if (key[Pad1[i]]) Stat&=~(1<<i);
	else Stat|=(1<<i);
    }

    pad->buttonStatus = Stat;
    pad->controllerType = 4; // standard

    return 0;
}

long CALLBACK PADreadPort2(PadDataS *pad) {
    static unsigned short Stat = 0xffff;
    int i;

    if (!(padflags & 2)) return 0xffff;

    for (i=0;i<16;i++) {
	if (key[Pad2[i]]) Stat&=~(1<<i);
	else Stat|=(1<<i);
    }

    pad->buttonStatus = Stat;
    pad->controllerType = 4; // standard

    return 0;
}

#if defined (__LINUX__)

long PADkeypressed(void) {
    int i,ret,retX;

    if (PadOpened == 0) return 0;

    ret = retX = 0;
    if (key[KEY_ESC]) { ret = KEY_ESC; retX = XK_Escape; }
    for (i=0;i<12;i++) if (key[i+KEY_F1]) { ret = KEY_F1+i; retX = XK_F1+i; }
    if (key[KEY_INSERT]) { ret = KEY_INSERT; retX = XK_Insert; }
    if (key[KEY_DEL]) { ret = KEY_DEL; retX = XK_Delete; }
    if (key[KEY_HOME]) { ret = KEY_HOME; retX = XK_Home; }
    if (key[KEY_END]) { ret = KEY_END; retX = XK_End; }
    if (key[KEY_PGUP]) { ret = KEY_PGUP; retX = XK_Page_Up; }
    if (key[KEY_PGDN]) { ret = KEY_PGDN; retX = XK_Page_Down; }

    if (!ret) return 0;

    for (i=0;i<16;i++) if (ret == Pad1[i]) return 0;
    for (i=0;i<16;i++) if (ret == Pad2[i]) return 0;

    key[ret] = 0;    
    return retX;
}

GtkWidget *ConfDlg;
GtkWidget *Edit;

void OnOk() {
    gtk_widget_hide(ConfDlg);
    SaveConf();
}

void OnCancel() {
    gtk_widget_hide(ConfDlg);
}

GtkWidget *CreateButton(GtkWidget *BBox, char *name) {
    GtkWidget *But = gtk_button_new_with_label(name);
    gtk_container_add(GTK_CONTAINER(BBox), But);
    GTK_WIDGET_SET_FLAGS(But, GTK_CAN_DEFAULT);
    gtk_widget_show(But);
    return But;
}

void OnKey(GtkWidget *But, gpointer data) {
    short *key = (short*)gtk_object_get_user_data(GTK_OBJECT(But));

    allegro_init();
    set_gfx_mode(GFX_SAFE, 320, 200, 0, 0);
    install_keyboard();
    textout(screen, font, "press any key", 100, 90, 0xffff);
    *key = readkey() >> 8;
    if (*key == KEY_ESC) *key = 0;
    remove_keyboard();
    allegro_exit();
}

#define CreateKeyBut(name, key) \
    Ok = CreateButton(BBox, name); \
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnKey), NULL); \
    gtk_object_set_user_data(GTK_OBJECT(Ok), (char *)&key);

long PADconfigure(void) {
    GtkWidget *Label;
    GtkWidget *Ok,*Cancel;
    GtkWidget *Box,*Box1,*BBox;

    LoadConf();

    ConfDlg = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(ConfDlg, 460, 320);
    gtk_window_set_title(GTK_WINDOW(ConfDlg), "PAD Config Dialog");
    gtk_window_set_position(GTK_WINDOW(ConfDlg), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(ConfDlg), 10);

    Box = gtk_vbox_new(0, 0);
    gtk_container_add(GTK_CONTAINER(ConfDlg), Box);
    gtk_widget_show(Box);

    Box1 = gtk_hbox_new(0, 0);
    gtk_box_pack_start(GTK_BOX(Box), Box1, FALSE, FALSE, 0);
    gtk_widget_show(Box1);

    Label = gtk_label_new("Pad1:");
    gtk_box_pack_start(GTK_BOX(Box1), Label, FALSE, FALSE, 0);
    gtk_widget_show(Label);

    BBox = gtk_vbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box1), BBox, FALSE, FALSE, 0);
    gtk_vbutton_box_set_spacing_default(0);
    gtk_widget_show(BBox);

    CreateKeyBut("Select", Pad1[0]);
    CreateKeyBut("Start", Pad1[3]);
    CreateKeyBut("Up", Pad1[4]);
    CreateKeyBut("Right", Pad1[5]);
    CreateKeyBut("Down", Pad1[6]);
    CreateKeyBut("Left", Pad1[7]);

    BBox = gtk_vbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box1), BBox, FALSE, FALSE, 0);
    gtk_vbutton_box_set_spacing_default(0);
    gtk_widget_show(BBox);

    CreateKeyBut("L2", Pad1[8]);
    CreateKeyBut("R2", Pad1[9]);
    CreateKeyBut("L1", Pad1[10]);
    CreateKeyBut("R1", Pad1[11]);
    CreateKeyBut("Triangle", Pad1[12]);
    CreateKeyBut("Circle", Pad1[13]);
    CreateKeyBut("Cross", Pad1[14]);
    CreateKeyBut("Square", Pad1[15]);

    Label = gtk_label_new("Pad2:");
    gtk_box_pack_start(GTK_BOX(Box1), Label, FALSE, FALSE, 10);
    gtk_widget_show(Label);

    BBox = gtk_vbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box1), BBox, FALSE, FALSE, 0);
    gtk_vbutton_box_set_spacing_default(0);
    gtk_widget_show(BBox);

    CreateKeyBut("Select", Pad2[0]);
    CreateKeyBut("Start", Pad2[3]);
    CreateKeyBut("Up", Pad2[4]);
    CreateKeyBut("Right", Pad2[5]);
    CreateKeyBut("Down", Pad2[6]);
    CreateKeyBut("Left", Pad2[7]);

    BBox = gtk_vbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box1), BBox, FALSE, FALSE, 0);
    gtk_vbutton_box_set_spacing_default(0);
    gtk_widget_show(BBox);

    CreateKeyBut("L2", Pad2[8]);
    CreateKeyBut("R2", Pad2[9]);
    CreateKeyBut("L1", Pad2[10]);
    CreateKeyBut("R1", Pad2[11]);
    CreateKeyBut("Triangle", Pad2[12]);
    CreateKeyBut("Circle", Pad2[13]);
    CreateKeyBut("Cross", Pad2[14]);
    CreateKeyBut("Square", Pad2[15]);

    BBox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box), BBox, FALSE, FALSE, 0);
    gtk_widget_show(BBox);

    Ok = CreateButton(BBox, "Ok");
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnOk), NULL);

    Cancel = CreateButton(BBox, "Cancel");
    gtk_signal_connect(GTK_OBJECT(Cancel), "clicked", GTK_SIGNAL_FUNC(OnCancel), NULL);

    gtk_widget_show(ConfDlg);

    return 0;
}

GtkWidget *AboutDlg;

void OnAboutOk() {
    gtk_widget_hide(AboutDlg);
}

void PADabout(void) {
    GtkWidget *Label;
    GtkWidget *Ok;
    GtkWidget *Box,*BBox;
    char AboutText[255];

    sprintf (AboutText,"%s %d.%d\n",libraryName, revision, build);

    AboutDlg = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(AboutDlg, 260, 80);
    gtk_window_set_title(GTK_WINDOW(AboutDlg), "PAD About Dialog");
    gtk_window_set_position(GTK_WINDOW(AboutDlg), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(AboutDlg), 10);

    Box = gtk_vbox_new(0, 0);
    gtk_container_add(GTK_CONTAINER(AboutDlg), Box);
    gtk_widget_show(Box);

    Label = gtk_label_new(AboutText);
    gtk_box_pack_start(GTK_BOX(Box), Label, FALSE, FALSE, 0);
    gtk_widget_show(Label);

    BBox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box), BBox, FALSE, FALSE, 0);
    gtk_widget_show(BBox);

    Ok = gtk_button_new_with_label("Ok");
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnAboutOk), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Ok);
    GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
    gtk_widget_show(Ok);

    gtk_widget_show(AboutDlg);
}

#elif defined(__WIN32__)

long CALLBACK PADconfigure(void) {
    return 0;
}

void CALLBACK PADabout(void) {
} 

#endif

long CALLBACK PADtest(void) {
    return 0;
}

void LoadConf() {
#if defined (__LINUX__)
    FILE *f;
    char cfg[255];

    sprintf (cfg,"%s/padAlleg.cfg",getenv("HOME"));
    f = fopen(cfg,"r");
    if (f == NULL) return;
    fread(Pad1, 2, 16, f);
    fread(Pad2, 2, 16, f);
    fclose(f);
#elif defined (__WIN32__)
#endif
}

void SaveConf() {
#ifdef __LINUX__
    FILE *f;
    char cfg[255];

    sprintf (cfg,"%s/padAlleg.cfg",getenv("HOME"));
    f = fopen(cfg,"w");
    if (f == NULL) return;
    fwrite(Pad1, 2, 16, f);
    fwrite(Pad2, 2, 16, f);
    fclose(f);
#endif
}
